from bs4 import BeautifulSoup
from datetime import datetime
import requests
from time import sleep


def get_arxiv_page(query: str,
                   baseURL: str = "http://export.arxiv.org/api/query?",
                   start: int = 0,
                   max_results: int = 10,
                   sortBy: str = "relevance",
                   sortOrder: str = "descending",
                   columns: list = [],
                   timeout: float = 10.) -> list:
    """
    Function processes the query and returns a list of data rows.

    Parameters
    ----------
    query : str
        Query to be requested
    baseURL : str, optional, default : "http://export.arxiv.org/api/query?"
        Base URL of the arxiv API
    start : int, optional, default : 0
        Starting index for page
    max_results : int, optional, default : 10
        Maximum number of page entries
    sortBy : str, optional, default : "relevance"
        Sort entries by
    sortOrder : str, optional, default : "descending"
        Order of sorting
    columns : list, optional, default : []
        List of columns to be returned
    timeout : float, optional, default : 10.
        Timeout in seconds for HTTP requests

    Returns
    -------
    rows : list
        List of data rows
    """

    # Check if columns are valid
    valid_columns = ["id", "title", "summary", "authors", "primary_category",
                     "categories", "comments", "updated", "published", "doi", "links"]
    for c in columns:
        if c not in valid_columns:
            raise ValueError(
                "{:s} is not a valid column name.\nValid column names are {}.".format(c, valid_columns))
    # Renaming map of columns for arxiv query
    col_map = {
        "id": "id",
        "title": "title",
        "summary": "summary",
        "authors": "author",
        "primary_category": "arxiv:primary_category",
        "categories": "category",
        "comments": "arxiv:comment",
        "updated": "updated",
        "published": "published",
        "doi": "arxiv:doi",
        "links": "link",
    }

    # Build and request query URL
    url = '{:s}{:s}&start={:d}&max_results={:d}&sortBy={:s}&sortOrder={:s}'.format(
        baseURL,
        query,
        start,
        max_results,
        sortBy,
        sortOrder
    )

    # Retry on server errors or timeouts
    retries = requests.adapters.Retry(total=5, backoff_factor=0.5,
                                      status_forcelist=[429, 500, 502, 503, 504])
    adapter = requests.adapters.HTTPAdapter(max_retries=retries)

    http = requests.Session()
    http.mount("http://", adapter)

    # The arXiv API can return an empty data set, even though there should be more pages.
    # Since it does not return an error code, we try again after one second.
    # Only after 5 tries with empty data sets, we assume that the query is done.
    i = 0
    while i < 5:
        response = http.get(url, timeout=timeout)

        # Read data and get entries
        data = BeautifulSoup(response.text, "xml")
        entries = data.find_all("entry")
        if entries != []:
            break
        sleep(0.5*2**i)
        i += 1

    # Loop over entries and build rows of data frame
    rows = []
    for entry in entries:

        d = {}

        # Parse the requested columns
        for c in columns:
            if c in ["authors", "categories", "primary_category", "links"]:
                tmp = entry.find_all(col_map[c])
                if tmp is not None:
                    v = []
                    for t in tmp:
                        if c == "authors":
                            v.append(t.find("name").text.strip())
                        elif c in ["categories", "primary_category"]:
                            v.append(t["term"].strip())
                        elif c == "links":
                            v.append(t["href"].strip())
                    val = "; ".join(v)
            else:
                tmp = entry.find(col_map[c])
                tmp_stripped = tmp.text.strip() if tmp is not None else ""
                if c == "id":
                    val = tmp_stripped[21:]
                elif c == "summary":
                    val = tmp_stripped.replace("\n", " ")
                elif c in ["published", "updated"]:
                    val = datetime.strptime(tmp_stripped, "%Y-%m-%dT%H:%M:%SZ")
                else:
                    val = tmp_stripped
            d[c] = val

        rows.append(d)

    return rows



from arxivloader.api import get_arxiv_page
import numpy as np
import pandas as pd
from time import sleep
from tqdm.auto import tqdm
import warnings


def load(query: str,
         num: int = 10,
         start: int = 0,
         page_size: int = 10,
         delay: float = 3.,
         sortBy: str = "relevance",
         sortOrder: str = "descending",
         columns: list = ["id", "title", "summary", "authors", "primary_category",
                          "categories", "comments", "updated", "published", "doi", "links"],
         timeout: float = 10.) -> pd.DataFrame:
    """
    Function returns a Pandas DataFrame with arXiv data for the given query.
    Please see the arXiv API documentation to build a valid query:
    https://arxiv.org/help/api/user-manual#_query_interface

    Please be responsible and abide by the arXiv API Terms of Usage:
    https://arxiv.org/help/api/tou

    Thank you to arXiv for use of its open access interoperability.

    Parameters
    ----------
    query : str
        Query to be requested
    num : int, optional, default : 10
        Number of entries to be returned
    start : int, optional, default : 0
        Starting index for page
    page_size : int, optional, defaul : 10
        Maximum number of page entries. Maximum is 30000.
    delay : float, optional, default : 3.
        Delay in seconds between page requests. Has to at least 3.
    sortBy : str, optional, default : "relevance"
        Sort entries by
    sortOrder : str, optional, default : "descending"
        Order of sorting
    columns : list, optional, default : ["id", "title", "summary", "authors", "primary_category", "categories", "comments", "updated", "published", "doi", "links"]
        Data columns to be retrieved
    timeout : float, optional, default : 10.
        Timeout in seconds for HTTP requests

    Returns
    -------
    df : pandas.DataFrame
        Pandas DataFrame with the requested arXiv data
    """

    if query == "":
        raise ValueError("Positional argument query is empty string.")

    if delay < 3:
        warnings.warn("Delay has to be at least 3 seconds. Setting delay=3.")
        delay = 3.

    if page_size > 30_000:
        warnings.warn("Maximum page size is 30000. Setting page_size=30000.")
        page_size = 30_000

    if sortBy not in ["relevance", "lastUpdatedDate", "submittedDate"]:
        raise ValueError(
            "Keyword argument sortBy can only be 'relevance', 'lastUpdatedDate', or 'submitedDate'.")

    if sortOrder not in ["ascending", "descending"]:
        raise ValueError(
            "Keyword argument sortOrder can only be 'ascending' or 'descending'.")

    start_ids = np.arange(start, start+num, page_size)

    rows = []
    start_ids = np.arange(start, start+num, page_size)
    is_empty = False
    with tqdm(start_ids, desc="Downloading pages") as iterator:
        for start_id in iterator:
            if not is_empty:
                if start_id > start_ids[0]:
                    sleep(delay)
                maxres = np.minimum(page_size, start+num-start_id)
                r = get_arxiv_page(
                    query, start=start_id, max_results=maxres, sortBy=sortBy, sortOrder=sortOrder, columns=columns, timeout=timeout)
                if r == []:
                    is_empty = True
                    continue
                rows += r

    df = pd.DataFrame(rows, columns=columns)

    # Dropping duplicate rows
    df.drop_duplicates(inplace=True, ignore_index=True)

    # Printing success message
    msg = "Retrieved {N_rows} entries.".format(N_rows=len(df))
    print(msg)

    return df
